/*
 * Clock and power management. This exposes the combined functionality
 * provided by the X1600 and related SoCs. The power management
 * functionality could be exposed using a separate driver.
 *
 * Copyright (C) 2017, 2018, 2020, 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-x1600.h"
#include <math.h>
#include <stdio.h>



// Register locations.

enum Regs : unsigned
{
  Clock_control       = 0x000,  // CPCCR
  Low_power_control   = 0x004,  // LCR
  Clock_gate0         = 0x020,  // CLKGR0
  Clock_gate1         = 0x028,  // CLKGR1
  Sleep_control       = 0x024,  // OPCR (oscillator and power control)
  Clock_status        = 0x0d4,  // CPCSR
  Ddr_divider         = 0x02c,  // DDRCDR
  Mac_divider         = 0x054,  // MACCDR
  I2s_divider0        = 0x060,  // I2SCDR
  I2s_divider1        = 0x070,  // I2S1CDR
  Lcd_divider         = 0x064,  // LPCDR
  Msc_divider0        = 0x068,  // MSC0CDR
  Msc_divider1        = 0x0a4,  // MSC1CDR
  Sfc_divider         = 0x074,  // SFCCDR
  Ssi_divider         = 0x05c,  // SSICDR
  Cim_divider         = 0x078,  // CIMCDR
  Pwm_divider         = 0x06c,  // PWMCDR
  Can_divider0        = 0x0a0,  // CAN0CDR
  Can_divider1        = 0x0a8,  // CAN1CDR
  Cdbus_divider       = 0x0ac,  // CDBUSCDR
  Macphy0_divider     = 0x0e4,  // MPHY0C
  Cpm_interrupt       = 0x0b0,  // CPM_INTR
  Cpm_interrupt_en    = 0x0b4,  // CPM_INTRE
  Cpm_swi             = 0x0bc,  // CPM_SFTINT
  Ddr_gate            = 0x0d0,  // DRCG
  Cpm_scratch_prot    = 0x038,  // CPSPPR
  Cpm_scratch         = 0x034,  // CPSPR
  Usb_param_control0  = 0x03c,  // USBPCR
  Usb_reset_detect    = 0x040,  // USBRDT
  Usb_vbus_jitter     = 0x044,  // USBVBFIL
  Usb_param_control1  = 0x048,  // USBPCR1
  Pll_control         = 0x00c,  // CPPCR
  Pll_control_A       = 0x010,  // CPAPCR
  Pll_control_M       = 0x014,  // CPMPCR
  Pll_control_E       = 0x018,  // CPEPCR
  Pll_fraction_A      = 0x084,  // CPAPACR
  Pll_fraction_M      = 0x088,  // CPMPACR
  Pll_fraction_E      = 0x08c,  // CPEPACR
};

enum Clock_source_bits : unsigned
{
  // Clock_control

  Clock_source_main     = 30, // SEL_SRC (output to SCLK_A)
  Clock_source_cpu      = 28, // SEL_CPLL (output to CCLK)
  Clock_source_hclock0  = 26, // SEL_H0PLL (output to AHB0)
  Clock_source_hclock2  = 24, // SEL_H2PLL (output to AHB2)

  // Divider registers

  Clock_source_can0     = 30, // CA0CS
  Clock_source_can1     = 30, // CA1CS
  Clock_source_cdbus    = 30, // CDCS
  Clock_source_cim      = 30, // CIMPCS
  Clock_source_ddr      = 30, // DCS
  Clock_source_i2s      = 31, // I2PCS
  Clock_source_lcd      = 30, // LPCS
  Clock_source_mac      = 30, // MACPCS
  Clock_source_msc0     = 30, // MPCS
  Clock_source_msc1     = 30, // MPCS
  Clock_source_pwm      = 30, // PWMPCS
  Clock_source_sfc      = 30, // SFCS
  Clock_source_ssi      = 30, // SPCS
};

enum Clock_source_values : unsigned
{
  Source_mME_main       = 0,
  Source_mME_pll_M      = 1,
  Source_mME_pll_E      = 2,

  // Special value

  Source_mask           = 0x3,
};

enum Clock_gate_bits : unsigned
{
  // Clock_control

  Clock_gate_main       = 23, // GATE_SCLKA

  // Clock_gate0

  Clock_gate_ddr        = 31, // DDR
  Clock_gate_ahb0       = 29, // AHB0
  Clock_gate_apb0       = 28, // APB0
  Clock_gate_rtc        = 27, // RTC
  Clock_gate_aes        = 24, // AES
  Clock_gate_lcd_pixel  = 23, // LCD
  Clock_gate_cim        = 22, // CIM
  Clock_gate_dma        = 21, // PDMA
  Clock_gate_ost        = 20, // OST
  Clock_gate_ssi0       = 19, // SSI0
  Clock_gate_timer      = 18, // TCU
  Clock_gate_dtrng      = 17, // DTRNG
  Clock_gate_uart2      = 16, // UART2
  Clock_gate_uart1      = 15, // UART1
  Clock_gate_uart0      = 14, // UART0
  Clock_gate_sadc       = 13, // SADC
  Clock_gate_audio      = 11, // AUDIO
  Clock_gate_ssi_slv    = 10, // SSI_SLV
  Clock_gate_i2c1       = 8,  // I2C1
  Clock_gate_i2c0       = 7,  // I2C0
  Clock_gate_msc1       = 5,  // MSC1
  Clock_gate_msc0       = 4,  // MSC0
  Clock_gate_otg        = 3,  // OTG
  Clock_gate_sfc        = 2,  // SFC
  Clock_gate_efuse      = 1,  // EFUSE
  Clock_gate_nemc       = 0,  // NEMC

  // Clock_gate1

  Clock_gate_arb        = 30, // ARB
  Clock_gate_mipi_csi   = 28, // MIPI_CSI
  Clock_gate_intc       = 26, // INTC
  Clock_gate_gmac0      = 23, // GMAC0
  Clock_gate_uart3      = 16, // UART3
  Clock_gate_i2s0_tx    = 9,  // I2S0_dev_tclk
  Clock_gate_i2s0_rx    = 8,  // I2S0_dev_rclk
  Clock_gate_hash       = 6,  // HASH
  Clock_gate_pwm        = 5,  // PWM
  Clock_gate_cdbus      = 2,  // CDBUS
  Clock_gate_can1       = 1,  // CAN1
  Clock_gate_can0       = 0,  // CAN0
};

enum Clock_change_enable_bits : unsigned
{
  Clock_change_enable_cpu       = 22,
  Clock_change_enable_ahb0      = 21,
  Clock_change_enable_ahb2      = 20,
  Clock_change_enable_ddr       = 29,
  Clock_change_enable_mac       = 29,
  Clock_change_enable_i2s       = 29,
  Clock_change_enable_lcd       = 29,
  Clock_change_enable_msc0      = 29,
  Clock_change_enable_msc1      = 29,
  Clock_change_enable_sfc       = 29,
  Clock_change_enable_ssi       = 29,
  Clock_change_enable_cim       = 29,
  Clock_change_enable_pwm       = 29,
  Clock_change_enable_can0      = 29,
  Clock_change_enable_can1      = 29,
  Clock_change_enable_cdbus     = 29,
};

enum Clock_busy_bits : unsigned
{
  Clock_busy_cpu       = 0,
  Clock_busy_ddr       = 28,
  Clock_busy_mac       = 28,
  Clock_busy_lcd       = 28,
  Clock_busy_msc0      = 28,
  Clock_busy_msc1      = 28,
  Clock_busy_sfc       = 28,
  Clock_busy_ssi       = 28,
  Clock_busy_cim       = 28,
  Clock_busy_pwm       = 28,
  Clock_busy_can0      = 28,
  Clock_busy_can1      = 28,
  Clock_busy_cdbus     = 28,
};

enum Clock_divider_bits : unsigned
{
  Clock_divider_can0      = 0,  // CAN0CDR
  Clock_divider_can1      = 0,  // CAN1CDR
  Clock_divider_cdbus     = 0,  // CDBUSCDR
  Clock_divider_cim       = 0,  // CIMCDR
  Clock_divider_cpu       = 0,  // CDIV
  Clock_divider_ddr       = 0,  // DDRCDR
  Clock_divider_hclock0   = 8,  // H0DIV (fast AHB peripherals)
  Clock_divider_hclock2   = 12, // H2DIV (fast AHB peripherals)
  Clock_divider_l2cache   = 4,  // L2CDIV
  Clock_divider_lcd       = 0,  // LPCDR
  Clock_divider_mac       = 0,  // MACCDR
  Clock_divider_msc0      = 0,  // MSC0CDR
  Clock_divider_msc1      = 0,  // MSC1CDR
  Clock_divider_pclock    = 16, // PDIV (slow APB peripherals)
  Clock_divider_pwm       = 0,  // PWMCDR
  Clock_divider_sfc       = 0,  // SFCCDR
  Clock_divider_ssi       = 0,  // SSICDR
};

enum Pll_bits : unsigned
{
  // Pll_control_A, Pll_control_M, Pll_control_E

  Pll_multiplier        = 20, // xPLLM
  Pll_input_division    = 14, // xPLLN
  Pll_output_division1  = 11, // xPLLOD1
  Pll_output_division0  = 8,  // xPLLOD0
  Pll_stable            = 3,  // xPLL_ON
  Pll_enabled           = 0,  // xPLLEN
};

enum Pll_bypass_bits : unsigned
{
  Pll_bypass_A          = 30, // APLL_BP
  Pll_bypass_M          = 28, // MPLL_BP
  Pll_bypass_E          = 26, // EPLL_BP
};



// Register field abstraction.

class Field
{
  uint32_t reg;
  uint32_t mask;
  uint8_t bit;
  bool defined;

public:
  explicit Field()
  : defined(false)
  {
  }

  explicit Field(uint32_t reg, uint32_t mask, uint32_t bit)
  : reg(reg), mask(mask), bit(bit), defined(true)
  {
  }

  uint32_t get_field(Cpm_regs &regs);
  void set_field(Cpm_regs &regs, uint32_t value);
  bool is_defined() { return defined; }
};

// Undefined fields.

Field Source_undefined, Gate_undefined, Change_enable_undefined, Busy_undefined, Divider_undefined;



// Common clock abstraction.

class Clock_base
{
protected:

  // Clock sources and source selection.

  int num_inputs;
  enum Clock_identifiers *inputs;
  Field _source;

public:
  explicit Clock_base(int num_inputs = 0,
                      enum Clock_identifiers inputs[] = NULL,
                      Field source = Source_undefined)
  : num_inputs(num_inputs), inputs(inputs), _source(source)
  {
  }

  // Clock control.

  virtual int have_clock(Cpm_regs &regs);
  virtual void start_clock(Cpm_regs &regs);
  virtual void stop_clock(Cpm_regs &regs);

  // Clock divider.

  virtual uint32_t get_divider(Cpm_regs &regs);
  virtual void set_divider(Cpm_regs &regs, uint32_t division);

  // Clock source.

  virtual uint8_t get_source(Cpm_regs &regs);
  virtual void set_source(Cpm_regs &regs, uint8_t source);

  // Clock source frequency.

  virtual uint32_t get_source_frequency(Cpm_regs &regs);

  // Output frequency.

  virtual uint32_t get_frequency(Cpm_regs &regs);
};



// PLL descriptions.

class Pll : public Clock_base
{
  uint32_t control_reg;
  enum Pll_bypass_bits bypass_bit;

public:
  explicit Pll(int num_inputs, enum Clock_identifiers inputs[],
               uint32_t control_reg, enum Pll_bypass_bits bypass_bit)
  : Clock_base(num_inputs, inputs), control_reg(control_reg), bypass_bit(bypass_bit)
  {
  }

  // PLL_specific control.

  int have_pll(Cpm_regs &regs);
  int pll_enabled(Cpm_regs &regs);
  int pll_bypassed(Cpm_regs &regs);

  // Clock control.

  int have_clock(Cpm_regs &regs);
  void start_clock(Cpm_regs &regs);
  void stop_clock(Cpm_regs &regs);

  // General frequency modifiers.

  uint16_t get_multiplier(Cpm_regs &regs);
  void set_multiplier(Cpm_regs &regs, uint16_t multiplier);
  uint8_t get_input_division(Cpm_regs &regs);
  void set_input_division(Cpm_regs &regs, uint8_t divider);
  uint8_t get_output_division(Cpm_regs &regs);
  void set_output_division(Cpm_regs &regs, uint8_t divider);

  // PLL output frequency.

  uint32_t get_frequency(Cpm_regs &regs);

  // Other operations.

  void set_pll_parameters(Cpm_regs &regs, uint16_t multiplier,
                          uint8_t in_divider, uint8_t out_divider);
};



// Clock descriptions.

class Clock : public Clock_base
{
  Field _gate, _change_enable, _busy, _divider;

  // Clock control.

  void change_disable(Cpm_regs &regs);
  void change_enable(Cpm_regs &regs);
  void wait_busy(Cpm_regs &regs);

public:
  explicit Clock(int num_inputs = 0, enum Clock_identifiers inputs[] = NULL,
                 Field source = Source_undefined,
                 Field gate = Gate_undefined,
                 Field change_enable = Change_enable_undefined,
                 Field busy = Busy_undefined,
                 Field divider = Divider_undefined)
  : Clock_base(num_inputs, inputs, source),
    _gate(gate), _change_enable(change_enable), _busy(busy), _divider(divider)
  {
  }

  // Clock control.

  int have_clock(Cpm_regs &regs);
  void start_clock(Cpm_regs &regs);
  void stop_clock(Cpm_regs &regs);

  // Clock divider.

  uint32_t get_divider(Cpm_regs &regs);
  void set_divider(Cpm_regs &regs, uint32_t division);

  // Clock source.

  void set_source(Cpm_regs &regs, uint8_t source);
};



// Clock instances.

#define Clock_inputs(...) ((enum Clock_identifiers []) {__VA_ARGS__})

Clock clock_ahb2_apb(3, Clock_inputs(Clock_none, Clock_main, Clock_pll_M),
                     Field(Clock_control, 3, Clock_source_hclock2));

Clock clock_aic_bitclk;

Clock clock_aic_pclk;

Clock clock_can0(4, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E, Clock_external),
                 Field(Can_divider0, 3, Clock_source_can0),
                 Field(Clock_gate1, 1, Clock_gate_can0),
                 Field(Can_divider0, 1, Clock_change_enable_can0),
                 Field(Can_divider0, 1, Clock_busy_can0),
                 Field(Can_divider0, 0xff, Clock_divider_can0));

Clock clock_can1(4, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E, Clock_external),
                 Field(Can_divider1, 3, Clock_source_can1),
                 Field(Clock_gate1, 1, Clock_gate_can1),
                 Field(Can_divider1, 1, Clock_change_enable_can1),
                 Field(Can_divider1, 1, Clock_busy_can1),
                 Field(Can_divider1, 0xff, Clock_divider_can1));

Clock clock_cdbus(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                  Field(Cdbus_divider, 3, Clock_source_cdbus),
                  Field(Clock_gate1, 1, Clock_gate_cdbus),
                  Field(Cdbus_divider, 1, Clock_change_enable_cdbus),
                  Field(Cdbus_divider, 1, Clock_busy_cdbus),
                  Field(Cdbus_divider, 0xff, Clock_divider_cdbus));

Clock clock_cim(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                Field(Cim_divider, 3, Clock_source_cim),
                Field(Clock_gate0, 1, Clock_gate_cim),
                Field(Cim_divider, 1, Clock_change_enable_cim),
                Field(Cim_divider, 1, Clock_busy_cim),
                Field(Cim_divider, 0xff, Clock_divider_cim));

Clock clock_cpu(3, Clock_inputs(Clock_none, Clock_main, Clock_pll_M),
                Field(Clock_control, 3, Clock_source_cpu),
                Gate_undefined,
                Field(Clock_control, 1, Clock_change_enable_cpu),
                Field(Clock_status, 1, Clock_busy_cpu),
                Field(Clock_control, 0x0f, Clock_divider_cpu));

Clock clock_ddr(3, Clock_inputs(Clock_none, Clock_main, Clock_pll_M),
                Field(Ddr_divider, 3, Clock_source_ddr),
                Field(Clock_gate0, 1, Clock_gate_ddr),
                Field(Ddr_divider, 1, Clock_change_enable_ddr),
                Field(Ddr_divider, 1, Clock_busy_ddr),
                Field(Ddr_divider, 0x0f, Clock_divider_ddr));

Clock clock_dma(1, Clock_inputs(Clock_pclock),
                Source_undefined,
                Field(Clock_gate0, 1, Clock_gate_dma));

Clock clock_emac;

Clock clock_external;

Clock clock_hclock0(3, Clock_inputs(Clock_none, Clock_main, Clock_pll_M),
                    Field(Clock_control, 3, Clock_source_hclock0),
                    Field(Clock_gate0, 1, Clock_gate_ahb0),
                    Field(Clock_control, 1, Clock_change_enable_ahb0),
                    Busy_undefined,
                    Field(Clock_control, 0x0f, Clock_divider_hclock0));

Clock clock_hclock2(1, Clock_inputs(Clock_ahb2_apb),
                    Source_undefined,
                    Field(Clock_gate0, 1, Clock_gate_apb0),
                    Field(Clock_control, 1, Clock_change_enable_ahb2),
                    Busy_undefined,
                    Field(Clock_control, 0x0f, Clock_divider_hclock2));

Clock clock_hdmi;

Clock clock_i2c(1, Clock_inputs(Clock_pclock),
                Source_undefined,
                Field(Clock_gate0, 1, Clock_gate_i2c0));

Clock clock_i2c0(1, Clock_inputs(Clock_pclock),
                 Source_undefined,
                 Field(Clock_gate0, 1, Clock_gate_i2c0));

Clock clock_i2c1(1, Clock_inputs(Clock_pclock),
                 Source_undefined,
                 Field(Clock_gate0, 1, Clock_gate_i2c1));

Clock clock_i2s;

Clock clock_i2s0_rx(2, Clock_inputs(Clock_main, Clock_pll_E),
                    Field(I2s_divider0, 1, Clock_source_i2s),
                    Field(Clock_gate1, 1, Clock_gate_i2s0_rx),
                    Field(I2s_divider0, 1, Clock_change_enable_i2s));

Clock clock_i2s0_tx(2, Clock_inputs(Clock_main, Clock_pll_E),
                    Field(I2s_divider0, 1, Clock_source_i2s),
                    Field(Clock_gate1, 1, Clock_gate_i2s0_tx),
                    Field(I2s_divider0, 1, Clock_change_enable_i2s));

Clock clock_kbc;

Clock clock_lcd;

Clock clock_lcd_pixel(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                      Field(Lcd_divider, 3, Clock_source_lcd),
                      Field(Clock_gate0, 1, Clock_gate_lcd_pixel),
                      Field(Lcd_divider, 1, Clock_change_enable_lcd),
                      Field(Lcd_divider, 1, Clock_busy_lcd),
                      Field(Lcd_divider, 0xff, Clock_divider_lcd));

Clock clock_mac(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                Field(Mac_divider, 3, Clock_source_mac),
                Field(Clock_gate1, 1, Clock_gate_gmac0),
                Field(Mac_divider, 1, Clock_change_enable_mac),
                Field(Mac_divider, 1, Clock_busy_mac),
                Field(Mac_divider, 0xff, Clock_divider_mac));

Clock clock_main(3, Clock_inputs(Clock_none, Clock_external, Clock_pll_A),
                 Field(Clock_control, 3, Clock_source_main),
                 Field(Clock_control, 1, Clock_gate_main));

Clock clock_msc(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                Field(Msc_divider0, 3, Clock_source_msc0),
                Field(Clock_gate0, 1, Clock_gate_msc0),
                Field(Msc_divider0, 1, Clock_change_enable_msc0),
                Field(Msc_divider0, 1, Clock_busy_msc0),
                Field(Msc_divider0, 0xff, Clock_divider_msc0));

Clock clock_msc0(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                 Field(Msc_divider0, 3, Clock_source_msc0),
                 Field(Clock_gate0, 1, Clock_gate_msc0),
                 Field(Msc_divider0, 1, Clock_change_enable_msc0),
                 Field(Msc_divider0, 1, Clock_busy_msc0),
                 Field(Msc_divider0, 0xff, Clock_divider_msc0));

Clock clock_msc1(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                 Field(Msc_divider1, 3, Clock_source_msc1),
                 Field(Clock_gate0, 1, Clock_gate_msc1),
                 Field(Msc_divider1, 1, Clock_change_enable_msc1),
                 Field(Msc_divider1, 1, Clock_busy_msc1),
                 Field(Msc_divider1, 0xff, Clock_divider_msc1));

Clock clock_none;

Clock clock_pclock(1, Clock_inputs(Clock_ahb2_apb),
                   Source_undefined,
                   Field(Clock_gate0, 1, Clock_gate_apb0),
                   Change_enable_undefined,
                   Busy_undefined,
                   Field(Clock_control, 0x0f, Clock_divider_pclock));

Pll clock_pll_A(1, Clock_inputs(Clock_external),
                Pll_control_A, Pll_bypass_A);

Pll clock_pll_E(1, Clock_inputs(Clock_external),
                Pll_control_E, Pll_bypass_E);

Pll clock_pll_M(1, Clock_inputs(Clock_external),
                Pll_control_M, Pll_bypass_M);

Clock clock_pwm(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                Field(Pwm_divider, 3, Clock_source_pwm),
                Field(Clock_gate1, 1, Clock_gate_pwm),
                Field(Pwm_divider, 1, Clock_change_enable_pwm),
                Field(Pwm_divider, 1, Clock_busy_pwm),
                Field(Pwm_divider, 0x0f, Clock_divider_pwm));

Clock clock_pwm0(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                 Field(Pwm_divider, 3, Clock_source_pwm),
                 Field(Clock_gate1, 1, Clock_gate_pwm),
                 Field(Pwm_divider, 1, Clock_change_enable_pwm),
                 Field(Pwm_divider, 1, Clock_busy_pwm),
                 Field(Pwm_divider, 0x0f, Clock_divider_pwm));

Clock clock_pwm1;

Clock clock_scc;

Clock clock_sfc(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                Field(Sfc_divider, 3, Clock_source_sfc),
                Field(Clock_gate0, 1, Clock_gate_sfc),
                Field(Sfc_divider, 1, Clock_change_enable_sfc),
                Field(Sfc_divider, 1, Clock_busy_sfc),
                Field(Sfc_divider, 0xff, Clock_divider_sfc));

Clock clock_smb0;

Clock clock_smb1;

Clock clock_smb2;

Clock clock_smb3;

Clock clock_smb4;

Clock clock_ssi(3, Clock_inputs(Clock_main, Clock_pll_M, Clock_pll_E),
                Field(Ssi_divider, 3, Clock_source_ssi),
                Field(Clock_gate0, 1, Clock_gate_ssi0),
                Field(Ssi_divider, 1, Clock_change_enable_ssi),
                Field(Ssi_divider, 1, Clock_busy_ssi),
                Field(Ssi_divider, 0xff, Clock_divider_ssi));

Clock clock_timer(1, Clock_inputs(Clock_pclock),
                  Source_undefined,
                  Field(Clock_gate0, 1, Clock_gate_timer));

Clock clock_uart0(1, Clock_inputs(Clock_external),
                  Source_undefined,
                  Field(Clock_gate0, 1, Clock_gate_uart0));

Clock clock_uart1(1, Clock_inputs(Clock_external),
                  Source_undefined,
                  Field(Clock_gate0, 1, Clock_gate_uart1));

Clock clock_uart2(1, Clock_inputs(Clock_external),
                  Source_undefined,
                  Field(Clock_gate0, 1, Clock_gate_uart2));

Clock clock_uart3(1, Clock_inputs(Clock_external),
                  Source_undefined,
                  Field(Clock_gate1, 1, Clock_gate_uart3));

Clock clock_udc;

Clock clock_uhc;

Clock clock_uprt;



// Clock register.

static Clock_base *clocks[Clock_identifier_count] = {
  &clock_ahb2_apb,
  &clock_aic_bitclk,
  &clock_aic_pclk,
  &clock_can0,
  &clock_can1,
  &clock_cdbus,
  &clock_cim,
  &clock_cpu,
  &clock_ddr,
  &clock_dma,
  &clock_emac,
  &clock_external,
  &clock_hclock0,
  &clock_hclock2,
  &clock_hdmi,
  &clock_i2c,
  &clock_i2c0,
  &clock_i2c1,
  &clock_i2s,
  &clock_i2s0_rx,
  &clock_i2s0_tx,
  &clock_kbc,
  &clock_lcd,
  &clock_lcd_pixel,
  &clock_mac,
  &clock_main,
  &clock_msc,
  &clock_msc0,
  &clock_msc1,
  &clock_none,
  &clock_pclock,
  &clock_pll_A,
  &clock_pll_E,
  &clock_pll_M,
  &clock_pwm,
  &clock_pwm0,
  &clock_pwm1,
  &clock_scc,
  &clock_sfc,
  &clock_smb0,
  &clock_smb1,
  &clock_smb2,
  &clock_smb3,
  &clock_smb4,
  &clock_ssi,
  &clock_timer,
  &clock_uart0,
  &clock_uart1,
  &clock_uart2,
  &clock_uart3,
  &clock_udc,
  &clock_uhc,
  &clock_uprt,
};



// Register access.

Cpm_regs::Cpm_regs(l4_addr_t addr, uint32_t exclk_freq)
: exclk_freq(exclk_freq)
{
  _regs = new Hw::Mmio_register_block<32>(addr);
}

// Utility methods.

uint32_t
Cpm_regs::get_field(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return (_regs[reg] & (mask << shift)) >> shift;
}

void
Cpm_regs::set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value)
{
  _regs[reg] = (_regs[reg] & (~(mask << shift))) | ((mask & value) << shift);
}



// Field methods.

uint32_t
Field::get_field(Cpm_regs &regs)
{
  if (defined)
    return regs.get_field(reg, mask, bit);
  else
    return 0;
}

void
Field::set_field(Cpm_regs &regs, uint32_t value)
{
  if (defined)
    regs.set_field(reg, mask, bit, value);
}



// Clock control.

int
Clock_base::have_clock(Cpm_regs &regs)
{
  (void) regs;
  return true;
}

void
Clock_base::start_clock(Cpm_regs &regs)
{
  (void) regs;
}

void
Clock_base::stop_clock(Cpm_regs &regs)
{
  (void) regs;
}

// Default divider.

uint32_t
Clock_base::get_divider(Cpm_regs &regs)
{
  (void) regs;
  return 1;
}

void
Clock_base::set_divider(Cpm_regs &regs, uint32_t division)
{
  (void) regs;
  (void) division;
}

// Clock sources.

uint8_t
Clock_base::get_source(Cpm_regs &regs)
{
  if (_source.is_defined())
    return _source.get_field(regs);
  else
    return 0;
}

void
Clock_base::set_source(Cpm_regs &regs, uint8_t source)
{
  if (_source.is_defined())
    return;

  _source.set_field(regs, source);
}

// Clock source frequencies.

uint32_t
Clock_base::get_source_frequency(Cpm_regs &regs)
{
  // Return the external clock frequency without any input clock.

  if (num_inputs == 0)
    return regs.exclk_freq;

  // Clocks with one source yield that input frequency.

  else if (num_inputs == 1)
    return clocks[inputs[0]]->get_frequency(regs);

  // With multiple sources, obtain the selected source for the clock.

  uint8_t source = get_source(regs);

  // Return the frequency of the source.

  if (source < num_inputs)
    return clocks[inputs[source]]->get_frequency(regs);
  else
    return 0;
}

// Output clock frequencies.

uint32_t
Clock_base::get_frequency(Cpm_regs &regs)
{
  return get_source_frequency(regs) / get_divider(regs);
}



// PLL-specific control.

int
Pll::have_pll(Cpm_regs &regs)
{
  return regs.get_field(control_reg, 1, Pll_stable);
}

int
Pll::pll_enabled(Cpm_regs &regs)
{
  return regs.get_field(control_reg, 1, Pll_enabled);
}

int
Pll::pll_bypassed(Cpm_regs &regs)
{
  return regs.get_field(control_reg, 1, bypass_bit);
}

// Clock control.

int
Pll::have_clock(Cpm_regs &regs)
{
  return have_pll(regs) && pll_enabled(regs);
}

void
Pll::start_clock(Cpm_regs &regs)
{
  regs.set_field(control_reg, 1, Pll_enabled, 1);
  while (!have_pll(regs));
}

void
Pll::stop_clock(Cpm_regs &regs)
{
  regs.set_field(control_reg, 1, Pll_enabled, 0);
  while (have_pll(regs));
}

// Feedback (13-bit) multiplier.

uint16_t
Pll::get_multiplier(Cpm_regs &regs)
{
  return regs.get_field(control_reg, 0x1fff, Pll_multiplier) + 1;
}

void
Pll::set_multiplier(Cpm_regs &regs, uint16_t multiplier)
{
  regs.set_field(control_reg, 0x1fff, Pll_multiplier, multiplier - 1);
}

// Input (6-bit) divider.

uint8_t
Pll::get_input_division(Cpm_regs &regs)
{
  return regs.get_field(control_reg, 0x3f, Pll_input_division) + 1;
}

void
Pll::set_input_division(Cpm_regs &regs, uint8_t divider)
{
  regs.set_field(control_reg, 0x3f, Pll_input_division, divider - 1);
}

// Output (dual 3-bit) dividers.

uint8_t
Pll::get_output_division(Cpm_regs &regs)
{
  uint8_t d0 = regs.get_field(control_reg, 0x07, Pll_output_division0);
  uint8_t d1 = regs.get_field(control_reg, 0x07, Pll_output_division1);

  return d0 * d1;
}

void
Pll::set_output_division(Cpm_regs &regs, uint8_t divider)
{
  // Assert 1 as a minimum.
  // Divider 0 must be less than or equal to divider 1.

  uint8_t d0 = (uint8_t) floor(sqrt(divider ? divider : 1));
  uint8_t d1 = divider / d0;

  regs.set_field(control_reg, 0x07, Pll_output_division0, d0);
  regs.set_field(control_reg, 0x07, Pll_output_division1, d1);
}

uint32_t
Pll::get_frequency(Cpm_regs &regs)
{
  // Test for PLL enable and not PLL bypass.

  if (pll_enabled(regs))
  {
    if (!pll_bypassed(regs))
      return (get_source_frequency(regs) * get_multiplier(regs)) /
             (get_input_division(regs) * get_output_division(regs));
    else
      return get_source_frequency(regs);
  }
  else
    return 0;
}

void
Pll::set_pll_parameters(Cpm_regs &regs, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  set_multiplier(regs, multiplier);
  set_input_division(regs, in_divider);
  set_output_division(regs, out_divider);

  if (pll_enabled(regs) && !pll_bypassed(regs))
    while (!have_pll(regs));
}



// Clock control.

void
Clock::change_disable(Cpm_regs &regs)
{
  if (_change_enable.is_defined())
    _change_enable.set_field(regs, 0);
}

void
Clock::change_enable(Cpm_regs &regs)
{
  if (_change_enable.is_defined())
    _change_enable.set_field(regs, 1);
}

int
Clock::have_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    return !_gate.get_field(regs);
  else
    return true;
}

void
Clock::start_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    _gate.set_field(regs, 0);
}

void
Clock::stop_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    _gate.set_field(regs, 1);
}

void
Clock::wait_busy(Cpm_regs &regs)
{
  if (_busy.is_defined())
    while (_busy.get_field(regs));
}



// Clock dividers.

uint32_t
Clock::get_divider(Cpm_regs &regs)
{
  if (_divider.is_defined())
    return _divider.get_field(regs) + 1;
  else
    return 1;
}

void
Clock::set_divider(Cpm_regs &regs, uint32_t division)
{
  if (_divider.is_defined())
    return;

  change_enable(regs);
  _divider.set_field(regs, division - 1);
  wait_busy(regs);
  change_disable(regs);
}

void
Clock::set_source(Cpm_regs &regs, uint8_t source)
{
  change_enable(regs);
  Clock_base::set_source(regs, source);
  wait_busy(regs);
  change_disable(regs);
}



// If implemented as a Hw::Device, various properties would be
// initialised in the constructor and obtained from the device tree
// definitions.

Cpm_x1600_chip::Cpm_x1600_chip(l4_addr_t addr, uint32_t exclk_freq)
: _cpm_regs(addr, exclk_freq)
{
  // add_cid("cpm");
  // add_cid("cpm-x1600");
  // register_property("exclk_freq", &exclk_freq);
}

int
Cpm_x1600_chip::have_clock(enum Clock_identifiers clock)
{
  return clocks[clock]->have_clock(_cpm_regs);
}

void
Cpm_x1600_chip::start_clock(enum Clock_identifiers clock)
{
  clocks[clock]->start_clock(_cpm_regs);
}

void
Cpm_x1600_chip::stop_clock(enum Clock_identifiers clock)
{
  clocks[clock]->stop_clock(_cpm_regs);
}

uint32_t
Cpm_x1600_chip::get_divider(enum Clock_identifiers clock)
{
  return clocks[clock]->get_divider(_cpm_regs);
}

void
Cpm_x1600_chip::set_divider(enum Clock_identifiers clock, uint32_t division)
{
  clocks[clock]->set_divider(_cpm_regs, division);
}

uint8_t
Cpm_x1600_chip::get_source(enum Clock_identifiers clock)
{
  return clocks[clock]->get_source(_cpm_regs);
}

void
Cpm_x1600_chip::set_source(enum Clock_identifiers clock, uint8_t source)
{
  clocks[clock]->set_source(_cpm_regs, source);
}

uint32_t
Cpm_x1600_chip::get_source_frequency(enum Clock_identifiers clock)
{
  return clocks[clock]->get_source_frequency(_cpm_regs);
}

uint32_t
Cpm_x1600_chip::get_frequency(enum Clock_identifiers clock)
{
  return clocks[clock]->get_frequency(_cpm_regs);
}

void
Cpm_x1600_chip::set_frequency(enum Clock_identifiers clock, uint32_t frequency)
{
  switch (clock)
  {
    // The pixel frequency is based on the selected clock source (SCLK_A, MPLL or
    // EPLL).

    case Clock_lcd_pixel:
    {

      // Switch to the MPLL and attempt to set the divider.

      Clock_base *lcd = clocks[Clock_lcd_pixel];
      Clock_base *pll = clocks[Clock_pll_M];

      lcd->set_source(_cpm_regs, Source_mME_pll_M);
      pll->start_clock(_cpm_regs);
      lcd->set_divider(_cpm_regs, lcd->get_source_frequency(_cpm_regs) / frequency);
      break;
    }

    default:
      break;
  }
}

void
Cpm_x1600_chip::set_pll_parameters(enum Clock_identifiers clock, uint16_t multiplier,
                                   uint8_t in_divider, uint8_t out_divider)
{
  Pll *pll = dynamic_cast<Pll *>(clocks[clock]);

  pll->set_pll_parameters(_cpm_regs, multiplier, in_divider, out_divider);
}



// C language interface functions.

void
*x1600_cpm_init(l4_addr_t cpm_base)
{
  /* Initialise the clock and power management peripheral with the
     register memory region and a 24MHz EXCLK frequency. */

  return (void *) new Cpm_x1600_chip(cpm_base, 24000000);
}

int
x1600_cpm_have_clock(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->have_clock(clock);
}

void
x1600_cpm_start_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->start_clock(clock);
}

void
x1600_cpm_stop_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_x1600_chip *>(cpm)->stop_clock(clock);
}

uint32_t
x1600_cpm_get_divider(void *cpm, enum Clock_identifiers clock)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_divider(clock);
}

void
x1600_cpm_set_divider(void *cpm, enum Clock_identifiers clock, uint32_t divider)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->set_divider(clock, divider);
}

uint8_t
x1600_cpm_get_source(void *cpm, enum Clock_identifiers clock)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_source(clock);
}

void
x1600_cpm_set_source(void *cpm, enum Clock_identifiers clock, uint8_t source)
{       
  static_cast<Cpm_x1600_chip *>(cpm)->set_source(clock, source);
}

uint32_t
x1600_cpm_get_source_frequency(void *cpm, enum Clock_identifiers clock)
{       
  return static_cast<Cpm_x1600_chip *>(cpm)->get_source_frequency(clock);
}

uint32_t
x1600_cpm_get_frequency(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_x1600_chip *>(cpm)->get_frequency(clock);
}

void
x1600_cpm_set_frequency(void *cpm, enum Clock_identifiers clock, uint32_t frequency)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_frequency(clock, frequency);
}

void
x1600_cpm_set_mpll_parameters(void *cpm, uint16_t multiplier, uint8_t in_divider, uint8_t out_divider)
{
  static_cast<Cpm_x1600_chip *>(cpm)->set_pll_parameters(Clock_pll_M, multiplier, in_divider, out_divider);
}
